from collections.abc import AsyncIterable

from aioplus.internal.coercions import to_async_iterable
from aioplus.internal.typing import SupportsBool


async def aany(aiterable: AsyncIterable[SupportsBool], /) -> bool:
    """
    Return :obj:`True` if any element of the async iterable evaluates to :obj:`True`.

    Parameters
    ----------
    aiterable : AsyncIterable of SupportsBool
        An asynchronous iterable of objects supporting :meth:`object.__bool__`.

    Returns
    -------
    :class:`bool`
        :obj:`True` if any element evaluates to :obj:`True`.
        :obj:`False` if all elements evaluate to :obj:`False`, or if the iterable is empty.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import aany, arange
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     aiterable = (num % 2 == 0 async for num in arange(2304))
    >>>     flg = await aany(aiterable)
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    Notes
    -----
    - Short-circuits on the first object that evaluates to :obj:`True`.

    See Also
    --------
    :func:`any`
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")

    async for value in aiterable:
        if value:
            return True

    return False



from collections.abc import AsyncIterable
from typing import Any, TypeVar, overload

from aioplus.internal.coercions import to_async_iterable
from aioplus.internal.sentinels import Sentinel


T = TypeVar("T")
D = TypeVar("D")


@overload
async def afirst(aiterable: AsyncIterable[T], /) -> T: ...


@overload
async def afirst(aiterable: AsyncIterable[T], /, *, default: D) -> T | D: ...


async def afirst(aiterable: AsyncIterable[Any], /, *, default: Any = Sentinel) -> Any:
    """Return the first item of the ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        An asynchronous iterable to retrieve the item from.

    default : D, optional
        A default value to return if the iterable is empty.
        If not provided, :obj:`IndexError` will be raised if the iterable is empty.

    Returns
    -------
    T or D
        The first item of ``aiterable`` or the default value.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import afirst, arange
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     aiterable = arange(4, 23)
    >>>     num = await afirst(aiterable)
    >>>     print(f'aiterable[0] = {num}')
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")

    aiterator = aiter(aiterable)
    value = await anext(aiterator, default)

    if value is not Sentinel:
        return value

    detail = "afirst(): empty iterable"
    raise IndexError(detail) from None



from collections.abc import AsyncIterable
from typing import Any, SupportsIndex, TypeVar, overload

from aioplus.internal.aenumerate import aenumerate
from aioplus.internal.coercions import to_async_iterable, to_non_negative_int
from aioplus.internal.sentinels import Sentinel


T = TypeVar("T")
D = TypeVar("D")


@overload
async def anth(aiterable: AsyncIterable[T], /, *, n: SupportsIndex) -> T: ...


@overload
async def anth(aiterable: AsyncIterable[T], /, *, n: SupportsIndex, default: D) -> T | D: ...


async def anth(
    aiterable: AsyncIterable[Any],
    /,
    *,
    n: SupportsIndex,
    default: Any = Sentinel,
) -> Any:
    """Return the nth item or a default value.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        An asynchronous iterable to retrieve the nth item from.

    n : SupportsIndex
        The index of the item to retrieve, starting from 0.

    default : D, optional
        A default value to return if the nth item does not exist.
        If not provided, :obj:`IndexError` will be raised if the nth item is not found.

    Returns
    -------
    T or D
        The nth item or the default value.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import anth, arange
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     aiterable = arange(23)
    >>>     value = await anth(aiterable, n=4)
    >>>     print(f'value = {value}')
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")
    n = to_non_negative_int(n, variable_name="n")

    async for index, value in aenumerate(aiterable):
        if index == n:
            return value

    if default is not Sentinel:
        return default

    detail = "The iterable contains fewer than 'n' items"
    raise IndexError(detail)



from collections.abc import AsyncIterable
from typing import SupportsIndex, TypeVar

from aioplus.internal.aislice import aislice
from aioplus.internal.coercions import to_async_iterable, to_non_negative_int


T = TypeVar("T")


def ahead(aiterable: AsyncIterable[T], /, *, n: SupportsIndex) -> AsyncIterable[T]:
    """Return the first ``n`` items of the ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        An asynchronous iterable to retrieve items from.

    n : SupportsIndex
        The number of items to retrieve from the start.

    Returns
    -------
    AsyncIterable[T]
        An asynchronous iterable yielding the first ``n`` items of the ``aiterable``.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import ahead, arange
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     async for num in ahead(arange(23), n=4):
    >>>         print(num)
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    See Also
    --------
    :func:`itertools.islice`
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")
    n = to_non_negative_int(n, variable_name="n")

    return aislice(aiterable, n)



from collections.abc import AsyncIterable

from aioplus.internal.coercions import to_async_iterable
from aioplus.internal.typing import SupportsBool


async def aall(aiterable: AsyncIterable[SupportsBool], /) -> bool:
    """
    Return :obj:`True` if all elements of the async iterable evaluate to :obj:`True`.

    Parameters
    ----------
    aiterable : AsyncIterable of SupportsBool
        An asynchronous iterable of objects supporting :meth:`object.__bool__`.

    Returns
    -------
    :class:`bool`
        :obj:`True` if all elements evaluate to :obj:`True`, or if the iterable is empty.
        :obj:`False` if any element evaluates to :obj:`False`.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import aall, arange
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     aiterable = (num > 0 async for num in arange(2304))
    >>>     flg = await aall(aiterable)
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    Notes
    -----
    - Short-circuits on the first object that evaluates to :obj:`False`.

    See Also
    --------
    :func:`all`
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")

    async for value in aiterable:
        if not value:
            return False

    return True



from collections.abc import AsyncIterable
from typing import TypeVar

from aioplus.internal.awindowed import awindowed
from aioplus.internal.coercions import to_async_iterable


T = TypeVar("T")


def atriplewise(aiterable: AsyncIterable[T], /) -> AsyncIterable[tuple[T, T, T]]:
    """Return successive overlapping triplets taken from the input ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable of T
        An asynchronous iterable of elements to be triplet.

    Returns
    -------
    AsyncIterable of tuple[T, T, T]
        An asynchronous iterable yielding triplets of elements.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import arange, atriplewise
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     async for left, middle, right in atriplewise(arange(23)):
    >>>         print(f'triplet = ({left}, {middle}, {right})')
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")

    return awindowed(aiterable, n=3)



from collections.abc import AsyncIterable, Callable
from concurrent.futures import Executor
from typing import LiteralString, ParamSpec, SupportsIndex, TypeVar


P = ParamSpec("P")
R = TypeVar("R")

T1 = TypeVar("T1")
T2 = TypeVar("T2")


def to_int(obj: SupportsIndex, /, *, variable_name: LiteralString) -> int:
    """Cast `SupportsIndex` to `int`."""
    if not isinstance(obj, SupportsIndex):
        detail = f"'{variable_name}' must be 'SupportsIndex'"
        raise TypeError(detail)

    obj = obj.__index__()

    if not isinstance(obj, int):
        detail = f"'{variable_name}.__index__()' must be 'int'"
        raise TypeError(detail)

    return obj


def to_positive_int(obj: SupportsIndex, /, *, variable_name: LiteralString) -> int:
    """Cast `SupportsIndex` to `int`.

    Notes
    -----
    * Raises `ValueError` if `obj.__index__()` is negative or zero.
    """
    obj = to_int(obj, variable_name=variable_name)

    if obj <= 0:
        detail = f"'{variable_name}' must be positive"
        raise ValueError(detail)

    return obj


def to_non_negative_int(obj: SupportsIndex, /, *, variable_name: LiteralString) -> int:
    """Cast `SupportsIndex` to `int`.

    Notes
    -----
    * Raises `ValueError` if `obj.__index__()` is negative.
    """
    obj = to_int(obj, variable_name=variable_name)

    if obj < 0:
        detail = f"'{variable_name}' must be non-negative"
        raise ValueError(detail)

    return obj


def to_async_iterable(
    obj: AsyncIterable[T1],
    /,
    *,
    variable_name: LiteralString,
) -> AsyncIterable[T1]:
    """Cast `object` to `AsyncIterable`.

    Notes
    -----
    * Raises `TypeError` if `obj` is not `AsyncIterable`.
    """
    if not isinstance(obj, AsyncIterable):
        detail = f"'{variable_name}' must be 'AsyncIterable'"
        raise TypeError(detail)

    return obj


def to_callable(obj: Callable[P, R], /, *, variable_name: LiteralString) -> Callable[P, R]:
    """Cast `object` to `Callable`.

    Notes
    -----
    * Raises `TypeError` if `obj` is not `Callable`.
    """
    if not callable(obj):
        detail = f"'{variable_name}' must be callable"
        raise TypeError(detail)

    return obj


def to_executor(obj: Executor, /, *, variable_name: LiteralString) -> Executor:
    """Cast `object` to `Executor`.

    Notes
    -----
    * Raises `TypeError` if `obj` is not `Executor`.
    """
    if not isinstance(obj, Executor):
        detail = f"'{variable_name}' must be 'Executor'"
        raise TypeError(detail)

    return obj


def to_bool(obj: bool, /, *, variable_name: LiteralString) -> bool:  # noqa: FBT001
    """Cast `object` to `bool`.

    Notes
    -----
    * Raises `TypeError` if `obj` is not `bool`.
    """
    if not isinstance(obj, bool):
        detail = f"'{variable_name}' must be 'bool'"
        raise TypeError(detail)

    return obj


def to_pair(obj: tuple[T1, T2], /, *, variable_name: LiteralString) -> tuple[T1, T2]:
    """Cast `object` to two-element tuple.

    Notes
    -----
    * Raises `TypeError` if `obj` is not `tuple`;
    * Raises `ValueError` if `len(obj) != 2`.
    """
    if not isinstance(obj, tuple):
        detail = f"'{variable_name}' must be 'tuple'"
        raise TypeError(detail)

    if len(obj) != 2:
        detail = f"'len({variable_name})' must be '2'"
        raise ValueError(detail)

    return obj



from collections.abc import AsyncIterable
from typing import Any

from aioplus.internal.coercions import to_async_iterable


async def alen(aiterable: AsyncIterable[Any], /) -> int:
    """Return length of the iterable.

    Parameters
    ----------
    aiterable : AsyncIterable[Any]
        An asynchronous iterable of objects.

    Returns
    -------
    :class:`int`
        Length of the iterable.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import alen, arange
    >>>
    >>> async def main() -> None:
    >>>     aiterable = arange(2304)
    >>>     length = await alen(aiterable)
    >>>     print(f"len(aiterable) == {length}")
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    See Also
    --------
    :func:`len`
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")

    count = 0
    async for _ in aiterable:
        count += 1

    return count



from collections.abc import AsyncIterable
from typing import TypeVar

from aioplus.internal.awindowed import awindowed
from aioplus.internal.coercions import to_async_iterable


T = TypeVar("T")


def apairwise(aiterable: AsyncIterable[T], /) -> AsyncIterable[tuple[T, T]]:
    """Return successive overlapping pairs taken from the input ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable of T
        An asynchronous iterable of elements to be paired.

    Returns
    -------
    AsyncIterable of tuple[T, T]
        An asynchronous iterable yielding pairs of elements.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import apairwise, arange
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     async for left, right in apairwise(arange(23)):
    >>>         print(f'pair = ({left}, {right})')
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    See Also
    --------
    :func:`itertools.pairwise`
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")

    return awindowed(aiterable, n=2)



from asyncio import get_running_loop
from collections.abc import Awaitable, Callable
from concurrent.futures import Executor
from contextvars import copy_context
from functools import partial, wraps
from typing import ParamSpec, TypeVar

from aioplus.internal.coercions import to_callable, to_executor


ReturnT = TypeVar("ReturnT")
ParamsT = ParamSpec("ParamsT")


def awaitify(
    func: Callable[ParamsT, ReturnT],
    /,
    *,
    executor: Executor | None = None,
) -> Callable[ParamsT, Awaitable[ReturnT]]:
    """Make a function asynchronous.

    Parameters
    ----------
    func : Callable
        A callable to be wrapped for asynchronous execution.

    executor : Executor, optional
        An optional :class:`concurrent.futures.Executor` to run the function in. If :obj:`None`, the
        default executor is used (usually a thread pool).

    Returns
    -------
    Callable
        An asynchronous callable that, when awaited, runs the original function in the executor.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import awaitify
    >>>
    >>> def func(num: int) -> None:
    >>>     '''Print the number.'''
    >>>     print(f'Num: {num}')
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     afunc = awaitify(func)
    >>>     await afunc(num=2304)
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    See Also
    --------
    :meth:`asyncio.loop.run_in_executor`
    """
    func = to_callable(func, variable_name="func")

    if executor is not None:
        executor = to_executor(executor, variable_name="executor")

    @wraps(func)
    async def afunc(*args: ParamsT.args, **kwargs: ParamsT.kwargs) -> ReturnT:
        """Run the function asynchronously."""
        loop = get_running_loop()
        context = copy_context()
        execute = partial(context.run, func, *args, **kwargs)
        return await loop.run_in_executor(executor, execute)

    return afunc



import asyncio

from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import Self, SupportsIndex, TypeVar

from aioplus.internal.coercions import to_non_negative_int


T = TypeVar("T")


def arepeat(
    obj: T,
    /,
    *,
    times: SupportsIndex | None = None,
) -> AsyncIterable[T]:
    """Yield the same object repeatedly, either infinitely or a fixed number of times.

    Parameters
    ----------
    obj : T
        The object to yield repeatedly.

    times : int, optional
        Number of repetitions. Must be an object supporting :meth:`object.__index__`.
        If :obj:`None`, the object is yielded indefinitely.

    Returns
    -------
    AsyncIterable of T
        An asynchronous iterable yielding the same object multiple times.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import arepeat
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     async for num in arepeat(23, times=4):
    >>>         print(num)
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    See Also
    --------
    :func:`itertools.repeat`
    """
    if times is not None:
        times = to_non_negative_int(times, variable_name="times")

    return ArepeatIterable(obj, times)


@dataclass
class ArepeatIterable(AsyncIterable[T]):
    """An repeated asynchronous iterable."""

    obj: T
    times: int | None

    def __aiter__(self) -> AsyncIterator[T]:
        """Return an asynchronous iterator."""
        return ArepeatIterator(self.obj, self.times)


@dataclass
class ArepeatIterator(AsyncIterator[T]):
    """An repeated asynchronous iterator."""

    obj: T
    times: int | None

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._count: int = 0

    def __aiter__(self) -> Self:
        """Return an asynchronous iterator."""
        return self

    async def __anext__(self) -> T:
        """Return the next value."""
        if self.times is not None and self._count >= self.times:
            raise StopAsyncIteration

        self._count += 1

        # Move to the next coroutine!
        await asyncio.sleep(0.0)

        return self.obj



from collections import deque
from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import Literal, Self, TypeVar, overload

from aioplus.internal.coercions import to_async_iterable, to_positive_int


T = TypeVar("T")


@overload
def awindowed(aiterable: AsyncIterable[T], /, *, n: Literal[2]) -> AsyncIterable[tuple[T, T]]: ...


@overload
def awindowed(
    aiterable: AsyncIterable[T],
    /,
    *,
    n: Literal[3],
) -> AsyncIterable[tuple[T, T, T]]: ...


@overload
def awindowed(aiterable: AsyncIterable[T], /, *, n: int) -> AsyncIterable[tuple[T, ...]]: ...


def awindowed(aiterable: AsyncIterable[T], /, *, n: int) -> AsyncIterable[tuple[T, ...]]:
    """Return a sliding window of width ``n`` over the given ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable of T
        An asynchronous iterable of elements to be windowed.

    Returns
    -------
    AsyncIterable of tuple[T, ...]
        An asynchronous iterable yielding windows of elements.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import arange, awindowed
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     async for window in awindowed(arange(23), n=4):
    >>>         print(f'window = {window}')
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())
    """
    aiterable = to_async_iterable(aiterable, variable_name="aiterable")
    n = to_positive_int(n, variable_name="n")

    return AwindowedIterable(aiterable, n=n)


@dataclass
class AwindowedIterable(AsyncIterable[tuple[T, ...]]):
    """An asynchronous iterable that windows data."""

    aiterable: AsyncIterable[T]
    n: int

    def __aiter__(self) -> AsyncIterator[tuple[T, ...]]:
        """Return an asynchronous iterator."""
        aiterator = aiter(self.aiterable)
        return AwindowedIterator(aiterator, self.n)


@dataclass
class AwindowedIterator(AsyncIterator[tuple[T, ...]]):
    """An asynchronous iterator that windows data."""

    aiterator: AsyncIterator[T]
    n: int

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._window: deque[T] = deque(maxlen=self.n)
        self._initialized_flg: bool = False
        self._finished_flg: bool = False

    def __aiter__(self) -> Self:
        """Return an asynchronous iterator."""
        return self

    async def __anext__(self) -> tuple[T, ...]:
        """Return the next value."""
        if self._finished_flg:
            raise StopAsyncIteration

        if not self._initialized_flg:
            await self._initialize()

        try:
            value = await anext(self.aiterator)

        except Exception:
            self._finished_flg = True
            self._window.clear()
            raise

        self._window.append(value)
        return tuple(self._window)

    async def _initialize(self) -> None:
        """Initialize the iterator."""
        self._initialized_flg = True
        try:
            for _ in range(self.n - 1):
                value = await anext(self.aiterator)
                self._window.append(value)

        except Exception:
            self._finished_flg = True
            self._window.clear()
            raise



from collections.abc import AsyncIterable, Callable
from typing import Any, TypeAlias, TypeVar, overload

from aioplus.internal import coercions
from aioplus.internal.sentinels import Sentinel
from aioplus.internal.typing import SupportsDunderGT, SupportsDunderLT


T = TypeVar("T")
D = TypeVar("D")

SupportsRichComparison: TypeAlias = SupportsDunderLT[Any] | SupportsDunderGT[Any]
SupportsRichComparisonT = TypeVar("SupportsRichComparisonT", bound=SupportsRichComparison)


@overload
async def amax(
    iterable: AsyncIterable[SupportsRichComparisonT],
    /,
    *,
    key: None = None,
) -> SupportsRichComparisonT: ...


@overload
async def amax(
    iterable: AsyncIterable[T],
    /,
    *,
    key: Callable[[T], SupportsRichComparison],
) -> T: ...


@overload
async def amax(
    iterable: AsyncIterable[SupportsRichComparisonT],
    /,
    *,
    key: None = None,
    default: T,
) -> SupportsRichComparisonT | T: ...


@overload
async def amax(
    iterable: AsyncIterable[T],
    /,
    *,
    key: Callable[[T], SupportsRichComparison],
    default: D,
) -> T | D: ...


async def amax(
    aiterable: AsyncIterable[Any],
    /,
    *,
    key: Callable[[Any], Any] | None = None,
    default: Any = Sentinel,
) -> Any:
    """Return the largest item in ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        An asynchronous iterable of objects.

    key : Callable[[T], SupportsRichComparison], optional
        A function that extracts a comparison key from each element in the iterable.

    default : D, optional
        A default value to return if the iterable is empty.

    Returns
    -------
    T or D
        The largest item in the iterable or the default value.

    Examples
    --------
    >>> import asyncio
    >>>
    >>> from aioplus import amax, arange
    >>>
    >>> async def main() -> None:
    >>>     '''Run the program.'''
    >>>     aiterable = arange(23)
    >>>     largest = await amax(aiterable)
    >>>     print(f'max(aiterable) == {largest}')
    >>>
    >>> if __name__ == '__main__':
    >>>     asyncio.run(main())

    See Also
    --------
    :func:`max`
    """
    aiterable = coercions.to_async_iterable(aiterable, variable_name="aiterable")

    if key is not None:
        key = coercions.to_callable(key, variable_name="key")

    aiterator = aiter(aiterable)

    try:
        largest = await anext(aiterator)

    except StopAsyncIteration:
        if default is not Sentinel:
            return default

        detail = "amax(): empty iterable"
        raise ValueError(detail) from None

    async for value in aiterator:
        largest = max(largest, value, key=key)

    return largest
